/*--------------------------------------------------------------------------
**      anon.h                   Twine           
**              
**------------------------- < License Information > ------------------------
**
**  
**      This file is released under terms of an LGPL.  Readers of 
**  this file may modify or use this source code so long as they make all
**  modifications publically available.  Please refer to LICENSE 
**  for further details.
**  
**
**      The original Copyright is in 1999, by CodeWeavers, Inc.
**
**      
**------------------------ < File Content Description > --------------------
**
**  Module:     include/twine/anon.h
**      
**  Description:
**      This function provides a compatibility layer of definitions
**  in order to compensate for a lack of anonymous unions or anonymous
**  structs in a native compiler.
**
**  Input definitions:
**      HAVE_ANON_STRUCT        From configure
**      HAVE_ANON_UNION            Whether or not the compiler supports them
**      
**      USE_ANON_STRUCT         Set by user in the Makefile.
**      USE_ANON_UNION              Default (and preference) is to not
**                                  define these.  For code which
**                                  relies upon anon structs/unions,
**                                  these #defines can be turned on. 
**                                  If the compiler supports
**                                  anonymous unions and structs,
**                                  the code will just work.        
**
**  Resulting defintions:
**      ANON_STRUCT(sname)      Can be used to name structures/unions as
**      ANON_UNION(sname)       either anonymous (no name) or with a
**                              fixed name, depending on the compilers
**                              abilities.
**                              
**      EMULATE_ANON_FIELD(sname, fname)
**                              Can be used as a last resort as
**                              a way of polluting the name space
**                              to emulate anonymous unions.
**                              That is, 'fname' is defined to
**                              be sname.fname if anon structs/unions
**                              are not available.
**
**
**  Usage:
**      It is this authors conviction that anonymous structures
**  and unions are an abomination upon the earth, and that 
**  they should be avoided in all code that we control.  
**      However, there exists non Twine code (e.g. MFC) that
**  relies upon anonymous structures and unions.
**  
**  Samples:
**      We imagine the following different possible examples:
**      1.  Internal Twine headers.
**            Action: Don't use any anonymous structs/unions
**            
**      2.  Twine headers which provide MS compatibility
**              and the MS headers supply anonymous structs/unions.
**              (e.g. propsht.h).
**            Action: Declare each anonymous union/struct
**                    using the ANON_STRUCT(xxx) where xxx
**                    is a good name for the structure.
**              struct {
**                  union {
**                      int field1;
**                      int field2;
**                  } ANON_UNION(union_1);
**
**                  struct {
**                      int field3;
**                      int field4;
**                  } ANON_STRUCT(struct_1);
**
**              } mystruct;
**                    
**  
**      3.  Twine internal code using the structures declared above.
**            Action:  Use the good name (e.g. xxx) to reference
**                     all fields.  Do _not_ use the anon stuff.
**              struct mystruct fred;
**              fred.union_1.field1 = 3;
**              fred.struct_1.field3 = 5;
**                     
**
**      4.  Non-Twine source code relying upon anon unions (e.g. MFC)
**          A.  Compiler supports anon structs and unions
**            Action:  If compiler supports anon structs/unions,
**                     just define USE_ANON_STRUCT and USE_ANON_UNION
**                     in the makefile and it should work.
**                     
**          B.  Compiler doesn't support anon structs and unions
**              Action: Add a line for each field used by the 
**              source file which fails to compile:
**      
**              (Prior to #includes, or in Makefile:)              
**
**              #define USE_ANON_STRUCT
**              #define USE_ANON_UNION
**              
**              (At top of source file):
**              
**              #define field1 EMULATE_ANON_FIELD(union_1, field1)
**              #define field3 EMULATE_ANON_FIELD(struct_1, field1)
**                         
**              struct mystruct fred;
**              fred.field1 = 3;
**              fred.field3 = 5;
**          
**              
**
**------------------------- < Revision Information > -----------------------
** $Header: /twinecvs/src/include/twine/anon.h,v 1.2 2000/01/07 22:47:30 jwhite Exp $
**
**      Full Revision history at bottom of file
**      
**--------------------------------------------------------------------------*/


#if ! defined(ANON_H_INCLUDED)
#define ANON_H_INCLUDED

/*----------------------------------------------------------------------------
**  Basic names to use when declaring a (hopefully) anonymous struct
**      and/or union.
**--------------------------------------------------------------------------*/
#if defined(HAVE_ANON_STRUCT) && defined(USE_ANON_STRUCT)
    #define ANON_STRUCT(sname)
#else
    #define ANON_STRUCT(sname)              sname
#endif

#if defined(HAVE_ANON_UNION) && defined(USE_ANON_UNION)
    #define ANON_UNION(sname)
#else
    #define ANON_UNION(sname)               sname
#endif


/*----------------------------------------------------------------------------
**  Macros to give you a way to refer to a field name inside of
**      a union that may be anonymous.
**--------------------------------------------------------------------------*/
#if defined(HAVE_ANON_STRUCT) && defined(USE_ANON_STRUCT)
    #define EMULATE_ANON_S_FIELD(sname, fname)    fname
#else
    #define EMULATE_ANON_S_FIELD(sname, fname)    sname.fname
#endif

#if defined(HAVE_ANON_STRUCT) && defined(USE_ANON_STRUCT)
    #define EMULATE_ANON_U_FIELD(uname, fname)    fname
#else
    #define EMULATE_ANON_U_FIELD(uname, fname)    uname.fname
#endif



#endif // ANON_H_INCLUDED

/*------------------------- < Full Revision History > ----------------------
** $Log: anon.h,v $
** Revision 1.2  2000/01/07 22:47:30  jwhite
** Revisions to remimplement the new anon struct/union.
** Revised to support 64 bit pointer platforms.
** Restored wtypes.h to version lost in Rob's latest changes.
**
** Revision 1.1  1999/09/02 02:56:50  jwhite
** Created; provides compatibility for systems that don't support anonymous
** structs and/or unions.
**
**
**-------------------------------------------------------------------------*/


