/*--------------------------------------------------------------------------
**      Hotkey.c                 Twin           From:  Twin/scheduler
**
**              
**------------------------- < License Information > ------------------------
**
**      This file was originally a part of Willows TWIN.  Willows
**  TWIN was released under a Library GPL (LGPL).  This permits
**  redistribution of this source code, provided that the full
**  TWIN license is in effect, and provided that all modifications
**  to this source code are made publicly available.
**  Please refer to Willows software (www.willows.com) or
**  LICENSE for full information.
**  
**      Under Twine, this file is also protected by an LGPL.  Please
**  see LICENSE for full details on this license.
**  
**
**      
**------------------------ < File Content Description > --------------------
**
**  Module:	scheduler/Hotkey.c
**
**  Description:
**      
**
**  Functions defined:
**    
**------------------------- < Revision Information > -----------------------
** $Header: /twinecvs/src/scheduler/Hotkey.c,v 1.2 1999/07/13 20:17:37 mwalsh Exp $
**
**      Full Revision history at bottom of file
**      
**--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
**              E X T E R N A L     I N C L U D E S 
**--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
**              I N T E R N A L     I N C L U D E S
**--------------------------------------------------------------------------*/
#include "windows.h"
#include "kerndef.h"
#include "Hotkey.h"

typedef struct _HOTKEY
{
  HWND hwnd;
  int id;
  UINT fsModifiers;
  UINT vk;
  struct _HOTKEY *next;
} HOTKEY;

HOTKEY *pFirstHotKey = NULL;

/*
 * CheckHotKey: Determines whether a given key message is a hotkey message
 * 
 * If the message is for a window and a key that we have a hotkey for, send a
 * WM_HOTKEY message instead.  We return TRUE if we sent (or possibly didn't send...)
 * our own message and the WM_KEY{UP,DOWN} message shouldn't be sent.
 * 
 * Additionally, every time we get a WM_KEYDOWN message (even if it's a repeat key),
 * we should send a WM_HOTKEY message.  We can just ignore it if it's a WM_KEYUP message
 * (even though we don't want a WM_KEYUP message to be sent to the app).
 */
BOOL CheckHotKey(LPMSG lpMsg)
{
  HOTKEY *pHotKey = pFirstHotKey;
  HOTKEY *pFound = NULL;
  int nExtra = 0, nNew;

  while (pHotKey)
    {
      /* Is this hotkey our key? */
      if ((!pHotKey->hwnd || (lpMsg->hwnd == pHotKey->hwnd)) &&
	  (pHotKey->vk == lpMsg->wParam) &&
	  (pHotKey->fsModifiers & MOD_SHIFT   ? GetKeyState(VK_SHIFT)   & 0x8000 : TRUE) &&
	  (pHotKey->fsModifiers & MOD_CONTROL ? GetKeyState(VK_CONTROL) & 0x8000 : TRUE) &&
	  (pHotKey->fsModifiers & MOD_ALT     ? GetKeyState(VK_MENU)    & 0x8000 : TRUE))
	{
	  /* Okay, we found one, how specific is it? */
	  nNew = 0;
	  if (GetKeyState(VK_SHIFT) & 0x8000)
	    nNew++;
	  if (GetKeyState(VK_CONTROL) & 0x8000)
	    nNew++;
	  if (GetKeyState(VK_MENU) & 0x8000)
	    nNew++;
	  if (pHotKey->fsModifiers & MOD_SHIFT)
	    nNew--;
	  if (pHotKey->fsModifiers & MOD_CONTROL)
	    nNew--;
	  if (pHotKey->fsModifiers & MOD_ALT)
	    nNew--;

	  /* nNew now holds the number of modifier keys down that were not originally specified */
	  if (!pFound || nNew < nExtra)
	    {
	      nExtra = nNew;
	      pFound = pHotKey;
	    }
	}

      /* No, loop */
      pHotKey = pHotKey->next;
    }

  /* We didn't find anything, give up */
  if (!pFound)
    return FALSE;

  /* Ignore WM_KEYUP messages with hotkeys */
  if (lpMsg->message == WM_KEYUP || lpMsg->message == WM_SYSKEYUP)
    return TRUE;

  /* We _did_ find something, send a hotkey message */
  if (pFound->hwnd)
    SendMessage(pFound->hwnd, WM_HOTKEY, pFound->id,
		MAKELPARAM((((GetKeyState(VK_SHIFT)   & 0x8000) ? MOD_SHIFT   : 0) &
			    ((GetKeyState(VK_CONTROL) & 0x8000) ? MOD_CONTROL : 0) &
			    ((GetKeyState(VK_MENU)    & 0x8000) ? MOD_ALT     : 0)),
			   lpMsg->wParam));
  else
    /* PostThreadMessage(GetCurrentThreadId(), WM_HOTKEY, pHotKey->id, */
    PostAppMessage(GetCurrentTask(), WM_HOTKEY, pFound->id,
		   MAKELPARAM((((GetKeyState(VK_SHIFT)   & 0x8000) ? MOD_SHIFT   : 0) &
			       ((GetKeyState(VK_CONTROL) & 0x8000) ? MOD_CONTROL : 0) &
			       ((GetKeyState(VK_MENU)    & 0x8000) ? MOD_ALT     : 0)),
			      lpMsg->wParam));
  return TRUE;
}

BOOL WINAPI RegisterHotKey(HWND hwnd, int id, UINT fsModifiers, UINT vk)
{
  HOTKEY **ppNextHotKey, *pHotKey, *pNewHotKey;

  /* Check that this hot key doesn't conflict with any others */
  ppNextHotKey = &pFirstHotKey;
  pHotKey = pFirstHotKey;
  while (pHotKey)
    {
      ppNextHotKey = &pHotKey->next;
      if (pHotKey->id == id)
	return FALSE; /* Same hotkey identifier */
      if ((pHotKey->vk == vk) &&
	  (pHotKey->fsModifiers == fsModifiers))
	return FALSE; /* Same keystroke */
      pHotKey = pHotKey->next;
    }

  pNewHotKey = (HOTKEY *)WinMalloc(sizeof(HOTKEY));
  if (!pNewHotKey)
    return FALSE;
  pNewHotKey->hwnd = hwnd;
  pNewHotKey->id = id;
  pNewHotKey->fsModifiers = fsModifiers;
  pNewHotKey->vk = vk;
  pNewHotKey->next = NULL;
  *ppNextHotKey = pNewHotKey;

  return TRUE;
}

BOOL WINAPI UnregisterHotKey(HWND hwnd, int id)
{
  HOTKEY **ppNextHotKey, *pHotKey;

  ppNextHotKey = &pFirstHotKey;
  pHotKey = pFirstHotKey;
  while (pHotKey)
    {
      if ((pHotKey->hwnd == hwnd) &&
	  (pHotKey->id == id))
	{
	  *ppNextHotKey = pHotKey->next;
	  WinFree((char *)pHotKey);
	  return TRUE;
	}
      ppNextHotKey = &pHotKey->next;
      pHotKey = pHotKey->next;
    }

  return FALSE;
}

/*------------------------- < Full Revision History > ----------------------
** $Log: Hotkey.c,v $
** Revision 1.2  1999/07/13 20:17:37  mwalsh
** Updated comments to use standard header and footer.
**
**-------------------------------------------------------------------------*/